from collections import namedtuple

import numpy as np
import pandas as pd

from sklearn.base import RegressorMixin, ClassifierMixin, clone
from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.model_selection._search import BaseSearchCV
from sklearn.utils import check_array
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.pipeline import Pipeline

TrainData = namedtuple(
    "TrainData",
    ("condition", "X", "y")
)
TrainResult = namedtuple(
    "TrainResult",
    ("condition", "model", "exp_columns", "predicted", "residual")
)
ChangingModels = namedtuple(
    "ChangingModels",
    ("name", "condition", "model")
)


class CausalBasedSimulator:
    """
    Causal based simulator.

    Attributes
    ----------
    train_result_ : dict of string -> list of dict
        information about trained models.

    residual_ : pandas.DataFrame
        residuals of trained models.

    cat_map_ : dict of strig -> list of str
        information about categorical variables. the key of its dictioary is
        variable name, the value is the list of classes.

    simulated_data_ : pandas.DataFrame
        result of simulation.
    """

    def train(self, X, causal_graph, models=None):
        """
        Estimate functional relations between variables and variable
        distributions based on the training data ``X`` and the causal graph
        ``G``. The functional relations represents by
        sklearn.linear_model.LinearRegression if the object variable is
        numeric, and represents by sklearn.linear_model.LogisticRegression
        if the object variable is categorical by default. ``train_result_``
        and ``residual_`` will be exposed after executing train().

        Parameters
        ----------
        X : pandas.DataFrame
            Training data.

        causal_graph : array-like of shape (n_features, _features)
            Causal graph.

        models : dict of string -> object, default=None
            Dictionary about models of variables. Models are cloned internaly
            and are trained to infer functioal relations. Given instances of
            the model are cloned to estimate the functional relation between
            variables.

        Returns
        -------
        self : Object
        """

        # checking inputs
        data, cat_map = self._check_data(X)

        causal_graph = self._check_causal_graph(causal_graph, data.shape[1])

        train_models = self._check_models(
            models,
            data.index,
            data.columns,
            cat_map
        )

        # training
        train_results = self._train(
            data,
            cat_map,
            causal_graph,
            train_models
        )

        residual_df = {}
        for col_name in X.columns:
            if len(train_results[col_name]) > 0:
                series = self._concat_residuals(
                    train_results,
                    col_name,
                    data.index
                )
            else:
                # node with no parent nodes
                series = np.ones(data.shape[0]) * np.nan

            residual_df[col_name] = series

        residual_df = pd.DataFrame(residual_df)

        # setting attributes
        self.train_result_ = self._prep_to_expose_train(
            train_results,
            cat_map
        )
        self.residual_ = residual_df
        self.categorical_info_ = cat_map

        # setting information
        self._data = data
        self._train_results = train_results
        self._causal_graph = causal_graph
        self._causal_order = None

        return self

    def run(
        self,
        changing_exog=None,
        changing_models=None,
        shuffle_residual=False,
        random_state=None
    ):
        """
        Generate simulated data using trained models and the given
        causal graph with given exogenous data and models.
        Specifying environmental changes to ``changing_exog`` or
        specifiyig changes in fucitonal relation to ``change_models``
        effects simulated data.
        Residuals to simulate variables are shuffled using radom_state
        if ``shuffle_residual`` is True. ``simulated_data_`` will be
        expose after excuting train().

        Parameters
        ----------
        changing_exog : dict of string -> array-like, default=None
            Dictioary about exogeous variables which keys are variable
            names and values are data of variables. That variable name
            should be a one of column names of X and the length should
            be same as X.

        changing_model : list of dict, default=None
            List of the changing models which elements are dictionary. that
            keys should be name, condition, and model. For name and
            condition, refer to ``train_result_`` and set the values
            corresponding to the conditions you wish to change. For model,
            you must provide a trained machine learning instance.

        shuffle_residual : bool, default=True
            If True, residuals are shuffled.

        random_state : int, default=None
            If shuffle_residual is True, random_state is used as seed.

        Returns
        -------
        simulated_data : pandas.DataFrame
            simulated data.
        """

        if self._train_results is None:
            raise RuntimeError(
                "run() shall be executed after train() is executed"
            )

        # checking inputs
        changing_exog_df = self._check_changing_exog(
            changing_exog,
            self._data.index,
            self._data.columns,
            self.categorical_info_
        )
        changing_models2 = self._check_changing_models(
            changing_models,
            self.categorical_info_,
            self._train_results
        )

        # calculating causal_order if it has not been calculated
        if self._causal_order is not None:
            causal_order = self._causal_order
        else:
            causal_order = self._get_causal_order(self._causal_graph)

        simulated_df = self._simulate(
            self._data,
            self._causal_graph,
            causal_order,
            self.categorical_info_,
            changing_exog_df,
            changing_models2,
            self._train_results,
            shuffle_residual,
            random_state
        )

        # set attributes
        self.simulated_data_ = simulated_df

        # set information
        self._changing_exog = changing_exog_df
        self._changing_models = changing_models2
        if self._causal_order is None:
            self._causal_order = causal_order

        return self.simulated_data_

    def _check_data(self, X):
        if not isinstance(X, pd.DataFrame):
            raise RuntimeError("X shall be a pandas.DataFrame.")

        cat_map = {}
        for c, is_cat in zip(X.columns, X.dtypes.values == "category"):
            if not is_cat:
                continue

            cat_map[c] = np.unique(X[c]).tolist()
            X[c] = X[c].apply(lambda x: cat_map[c].index(x))

        return X, cat_map

    def _check_causal_graph(self, causal_graph, col_num):
        graph = check_array(causal_graph)

        if graph.shape[0] != col_num or graph.shape[1] != col_num:
            raise RuntimeError("causal_graph.shape shall be square.")

        return graph

    def _check_model_instance(self, model, col_name, cat_map):
        if col_name not in cat_map.keys():
            model_type = RegressorMixin
        else:
            model_type = ClassifierMixin

        if isinstance(model, Pipeline):
            if not isinstance(model.steps[-1][-1], model_type):
                raise RuntimeError(
                    "The last step in Pipeline should be an "
                    + "instance of a regression/classification model."
                )
        elif isinstance(model, BaseSearchCV):
            if not isinstance(model.get_params()["estimator"], model_type):
                raise RuntimeError(
                    "The type of the estimator shall be an "
                    + "instance of a regression/classification model."
                )
        else:
            if not isinstance(model, model_type):
                raise RuntimeError(
                    "The type of the estimator shall be an "
                    + "instance of a regression/classification model."
                )

        if model_type == ClassifierMixin:
            try:
                func = getattr(model, "predict_proba")
                if not callable(func):
                    raise Exception
            except Exception:
                raise RuntimeError(
                    "Classification models shall have "
                    + "predict_proba()."
                )

    def _check_models(self, models, index, columns, cat_map):
        if models is None:
            return {}

        if not isinstance(models, dict):
            raise RuntimeError("models shall be a dictionary.")

        for col_name, model in models.items():
            if col_name not in columns:
                raise RuntimeError(f"Unknown column name ({col_name})")

            self._check_model_instance(model, col_name, cat_map)

        return models

    def _check_changing_exog(self, changing_exog, index, columns, cat_map):
        if changing_exog is None:
            return pd.DataFrame()

        if not isinstance(changing_exog, dict):
            raise RuntimeError("changing_exog shall be a dictionary.")

        changing_exog_df = {}
        for col_name, values in changing_exog.items():
            if col_name not in columns:
                raise RuntimeError(
                    f"Unknown key in changing_exog. ({col_name})"
                )

            if col_name in cat_map.keys():
                raise RuntimeError(
                    f"Category variables shall not be specified. ({col_name})"
                )

            s = check_array(values, ensure_2d=False, dtype=None).flatten()
            if s.shape[0] != len(index):
                raise RuntimeError(
                    f"Wrong length. ({s.shape[0]} != {len(index)})"
                )

            changing_exog_df[col_name] = pd.Series(values, index=index)
        changing_exog_df = pd.DataFrame(changing_exog_df).loc[index, :]

        return changing_exog_df

    def _check_changing_models(self, changing_models, cat_map, train_results):
        if changing_models is None:
            return {}

        if not isinstance(changing_models, list):
            raise RuntimeError("changing_models shall be a list.")

        changing_models_ = []
        for model_info in changing_models:
            if not isinstance(model_info, dict):
                raise RuntimeError(
                    "changing_models shall be a list of dictionaries."
                )

            missing_keys = set(ChangingModels._fields) - set(model_info.keys())
            if len(missing_keys) > 0:
                raise RuntimeError(
                    "Missing key on model_info. " + str(missing_keys)
                )

            name = model_info["name"]
            if name not in train_results.keys():
                raise RuntimeError(f"Unknown name. ({name})")

            condition = model_info["condition"]
            conditions = [
                cond_result.condition for cond_result in train_results[name]
            ]
            if condition not in conditions:
                raise RuntimeError("Not-exsitent consition. " + str(condition))

            model = model_info["model"]
            self._check_model_instance(model, name, cat_map)

            changing_models_.append(
                ChangingModels(name=name, condition=condition, model=model)
            )
        return changing_models_

    def _get_causal_order(self, causal_graph):
        causal_order = []

        row_num = causal_graph.shape[0]
        original_index = np.arange(row_num)

        while 0 < len(causal_graph):
            # finding rows where the elements are all zeros
            row_index_list = np.where(
                np.sum(np.abs(causal_graph), axis=1) == 0
            )[0]
            if len(row_index_list) == 0:
                break

            target_index = row_index_list[0]

            # append i to the end of the list
            causal_order.append(original_index[target_index])
            original_index = np.delete(original_index, target_index, axis=0)

            # remove i-th row and i-th column from matrix
            mask = np.delete(
                np.arange(len(causal_graph)),
                target_index,
                axis=0
            )
            causal_graph = causal_graph[mask][:, mask]

        if len(causal_order) != row_num:
            causal_order = None

        return np.array(causal_order)

    def _get_train_data(
        self,
        data,
        X_cols,
        y_col,
        cat_map,
        changing_exog_df=None
    ):
        train_data_list = []

        # column names of categorical variables in X_columns
        X_cols_cat = set(X_cols) & set(cat_map.keys())
        X_cols_cat = sorted(X_cols_cat, key=lambda x: list(X_cols).index(x))

        # no categorical variable in X
        if len(X_cols_cat) == 0:
            X = data.loc[:, X_cols].copy()
            y = data.loc[:, y_col].copy()

            # overwrite with exogenous
            if changing_exog_df is not None:
                for col, series in changing_exog_df.items():
                    if col in X.columns.tolist():
                        X[col] = series

            train_data_list.append(TrainData(condition=None, X=X, y=y))
            return train_data_list

        # column names of non-categorical variables in X_columns
        X_cols_num = set(X_cols) - set(cat_map.keys())
        X_cols_num = sorted(X_cols_num, key=lambda x: list(X_cols).index(x))

        # unique conditions by categorical parent nodes
        uniq_conds = np.unique(
            data.loc[:, X_cols_cat].values.tolist(),
            axis=0
        )

        for uniq_cond in uniq_conds:
            # filter by the condition
            target_index = np.argwhere(
                np.all(data.loc[:, X_cols_cat].values == uniq_cond, axis=1)
            ).flatten()

            X_filtered = data.iloc[target_index, :].loc[:, X_cols_num].copy()
            y_filtered = data.iloc[target_index, :].loc[:, y_col].copy()

            # overwrite with exogenous
            if changing_exog_df is not None:
                for col, series in changing_exog_df.items():
                    if col in X.columns.tolist():
                        X_filtered[col] = series.loc[target_index]

            condition = {}
            for col_name, value in zip(X_cols_cat, uniq_cond):
                condition[col_name] = cat_map[col_name][value]

            train_data_list.append(
                TrainData(condition=condition, X=X_filtered, y=y_filtered)
            )

        return train_data_list

    def _predict_reg_model(self, X, model):
        try:
            predicted = model.predict(X.values)
        except Exception as e:
            raise RuntimeError(
                "An exception occurred during predict() of the model. "
                + str(e)
            )

        return np.array(predicted)

    def _train_reg_model(self, X, y, model):
        try:
            model.fit(X.values, y.values)
        except Exception as e:
            raise RuntimeError(
                "An exception occurred during fit() of the model. "
                + str(e)
            )

        predicted = self._predict_reg_model(X, model)
        predicted = pd.Series(predicted, index=X.index, dtype=y.dtype)

        resid = y - predicted

        return model, predicted, resid

    def _predict_clf_model(self, X, model):
        try:
            proba = model.predict_proba(X.values)
        except Exception as e:
            raise RuntimeError(
                "An exception occurred during predict_proba() of the model. "
                + str(e)
            )

        # sampling values based on predicted probability
        predicted = []
        for proba_ in proba:
            predicted.append(np.random.choice(model.classes_, p=proba_))

        return np.array(predicted)

    def _train_clf_model(self, X, y, model, cat_map):
        try:
            model.fit(X.values, y.values)
        except Exception as e:
            raise RuntimeError(
                "An exception occurred during fit() of the model. "
                + str(e)
            )

        predicted = self._predict_clf_model(X, model)
        predicted = pd.Series(predicted, index=X.index, dtype=y.dtype)

        # classifier doesn't have residuals
        resid = None
        return model, predicted, resid

    def _concat_residuals(self, train_results, to_name, index):
        ret = []

        cond_results = train_results[to_name]

        for cond_result in cond_results:
            series = getattr(cond_result, "residual")

            if series is None:
                predicted = getattr(cond_result, "predicted")
                series = pd.Series(
                    np.ones(predicted.shape[0]) * np.nan,
                    index=predicted.index
                )

            ret.append(series)

        # concat and sort by original index
        ret = pd.concat(ret, axis=0).loc[index]

        return ret

    def _concat_predicts(self, train_results, to_name, index):
        ret = []

        cond_results = train_results[to_name]
        for cond_result in cond_results:
            series = getattr(cond_result, "predicted")
            ret.append(series)
        ret = pd.concat(ret, axis=0).loc[index]

        return ret

    def _prep_to_expose_train(self, train_results, cat_map):
        train_results_ = {}

        for to_name, cond_results in train_results.items():
            train_result_ = []

            for result in cond_results:
                result_ = {}

                if result.condition is None:
                    result_["condition"] = None
                else:
                    result_["condition"] = {
                        k: v for k, v in result.condition.items()
                    }

                result_["model"] = result.model

                result_["exp_columns"] = result.exp_columns

                if to_name not in cat_map.keys():
                    # non categorical
                    result_["predicted"] = result.predicted
                else:
                    # categorical
                    result_["predicted"] = result.predicted.map(
                        lambda x: cat_map[to_name][x]
                    )

                result_["residual"] = result.residual

                train_result_.append(result_)
            train_results_[to_name] = train_result_

        return train_results_

    def _train(self, data, cat_map, graph, models):
        train_results = {}

        for to_index, graph_row in enumerate(graph):
            cond_results = []

            from_indices = np.argwhere(
                ~np.isclose(graph_row, 0)
            ).flatten()

            # get column names
            to_name = data.columns[to_index]
            from_names = data.columns[from_indices]
            from_names_num = sorted(
                set(from_names) - set(cat_map.keys()),
                key=lambda x: list(data.columns).index(x)
            )

            # making data list splitted by conditions
            cond_tr_data = self._get_train_data(
                data,
                from_names,
                to_name,
                cat_map
            )

            # node with no parents
            if len(from_names) == 0:
                train_results[to_name] = []
                continue

            # node with parents
            for tr_data in cond_tr_data:
                # selecting a model to train
                if to_name in models.keys():
                    model = clone(models[to_name])
                else:
                    if len(from_names_num) > 0:
                        # default
                        if to_name in cat_map.keys():
                            model = LogisticRegression()
                        else:
                            model = LinearRegression()
                    else:
                        # no non-categorical parents
                        if to_name in cat_map.keys():
                            model = DummyClassifier(strategy="uniform")
                        else:
                            model = DummyRegressor()

                # train model
                if to_name in cat_map.keys():
                    model, predicted, resid = self._train_clf_model(
                        tr_data.X,
                        tr_data.y,
                        model,
                        cat_map
                    )
                else:
                    model, predicted, resid = self._train_reg_model(
                        tr_data.X,
                        tr_data.y,
                        model
                    )

                cond_results.append(TrainResult(
                    condition=tr_data.condition,
                    model=model,
                    exp_columns=from_names_num,
                    predicted=predicted,
                    residual=resid
                ))
            train_results[to_name] = cond_results

        return train_results

    def _select_changing_model(
        self,
        changing_models,
        to_name,
        condition
    ):
        for model_info in changing_models:
            if model_info.name != to_name:
                continue
            elif model_info.condition != condition:
                continue

            return model_info.model
        return None

    def _simulate(
        self,
        data,
        causal_graph,
        causal_order,
        cat_map,
        changing_exog_df,
        changing_models,
        train_results,
        shuffle_residual,
        random_state
    ):
        sim_df = pd.DataFrame(index=data.index)

        # predicting from upstream to downstream
        for to_index in causal_order:
            # input nodes of node i are nonzero variables of i-th row of graph
            from_indices = causal_graph[to_index, :]
            from_indices = np.argwhere(~np.isclose(from_indices, 0)).flatten()

            to_name = data.columns[to_index]
            from_names = [
                data.columns[from_index] for from_index in from_indices
            ]

            # assigning values without predicting if to_name has no parents.
            if len(from_names) == 0:
                sim_df[to_name] = data[to_name].copy()
                continue

            if to_name in changing_exog_df.columns:
                error = changing_exog_df[to_name]
            elif to_name not in cat_map.keys():
                error = self._concat_residuals(
                    train_results,
                    to_name,
                    data.index
                )

            X = sim_df[from_names]

            # simulation for each conditions
            y_sim = []
            for cond_result in train_results[to_name]:
                X_ = X

                if cond_result.condition is not None:
                    # filter rows by a condition
                    filter_ = []
                    for key, value in cond_result.condition.items():
                        filter_.append(X_[key] == cat_map[key].index(value))
                    filter_ = np.all(filter_, axis=0)
                    X_ = X_.loc[filter_, :]

                # filter columns by explanatories
                if len(cond_result.exp_columns) > 0:
                    X_ = X_.loc[:, cond_result.exp_columns]
                else:
                    # dummy data when all parent nodes are categorical
                    X_ = pd.DataFrame(
                        np.ones(X_.shape[0]).reshape(X_.shape[0]) * np.nan,
                        index=X_.index,
                        columns=[to_name]
                    )

                # selecting a model for the simulation
                model = self._select_changing_model(
                    changing_models,
                    to_name,
                    cond_result.condition
                )
                if model is None:
                    model = cond_result.model

                if to_name in cat_map.keys():
                    y_hat = self._predict_clf_model(X_, model)
                else:
                    y_hat = self._predict_reg_model(X_, model)

                    cond_error = error.loc[X_.index]
                    if shuffle_residual is True:
                        cond_error = cond_error.sample(
                            frac=1,
                            random_state=random_state
                        )
                    y_hat += cond_error

                y_sim.append(pd.Series(y_hat, index=X_.index))
            sim_df[to_name] = pd.concat(y_sim, axis=0).loc[data.index]
        sim_df = sim_df.loc[data.index, data.columns]

        # decoding categorical values
        for col_name, categories in cat_map.items():
            sim_df[col_name] = sim_df[col_name].map(
                lambda idx: categories[idx]
            )
            sim_df[col_name] = sim_df[col_name].astype("category")

        return sim_df



import itertools

import numpy as np
from scipy.optimize import fmin_l_bfgs_b
from scipy.stats import pearsonr, shapiro
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array

from ..hsic import get_gram_matrix, get_kernel_width, hsic_test_gamma, hsic_teststat


def extract_ancestors(
    X,
    max_explanatory_num=2,
    cor_alpha=0.01,
    ind_alpha=0.01,
    shapiro_alpha=0.01,
    MLHSICR=True,
    bw_method="mdbs",
):
    """Extract a set of ancestors of each variable
    Implementation of RCD Algorithm1 [1]_

    References
    ----------
    .. [1] T.N.Maeda and S.Shimizu. RCD: Repetitive causal discovery of
        linear non-Gaussian acyclic models with latent confounders.
        In Proc. 23rd International Conference on Artificial Intelligence and
        Statistics (AISTATS2020), Palermo, Sicily, Italy. PMLR  108:735-745, 2020.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    max_explanatory_num : int, optional (default=2)
        Maximum number of explanatory variables.
    cor_alpha : float, optional (default=0.01)
        Alpha level for pearson correlation.
    ind_alpha : float, optional (default=0.01)
        Alpha level for HSIC.
    shapiro_alpha : float, optional (default=0.01)
        Alpha level for Shapiro-Wilk test.
    MLHSICR : bool, optional (default=False)
        If True, use MLHSICR for multiple regression, if False, use OLS for multiple regression.
    bw_method : str, optional (default=``mdbs``)
            The method used to calculate the bandwidth of the HSIC.

        * ``mdbs`` : Median distance between samples.
        * ``scott`` : Scott's Rule of Thumb.
        * ``silverman`` : Silverman's Rule of Thumb.
    """
    n_features = X.shape[1]
    M = [set() for i in range(n_features)]
    l = 1
    hu_history = {}

    X = check_array(X)

    while True:
        changed = False
        U_list = itertools.combinations(range(n_features), l + 1)
        U_list = [sorted(list(U)) for U in U_list]

        for U in U_list:
            # Get the set of common ancestors of U
            H_U = get_common_ancestors(M, U)

            if tuple(U) in hu_history and H_U == hu_history[tuple(U)]:
                continue

            Y = get_residual_matrix(X, U, H_U)

            # Shapiro-Wilk test
            if not is_non_gaussianity(Y, U, shapiro_alpha):
                continue

            # Pearson's correlation
            is_cor = True
            for xi, xj in itertools.combinations(U, 2):
                if not is_correlated(Y[:, xi], Y[:, xj], cor_alpha):
                    is_cor = False
                    break
            if not is_cor:
                continue

            sink_set = []
            for xi in U:
                xj_list = list(set(U) - set([xi]))

                if exists_ancestor_in_U(M, U, xi, xj_list):
                    continue

                # Check whether the residuals obtained from multiple regressions are independent
                if is_independent_of_resid(
                    Y, xi, xj_list, ind_alpha, bw_method, MLHSICR
                ):
                    sink_set.append(xi)

            if len(sink_set) == 1:
                xi = sink_set[0]
                xj_list = list(set(U) - set(sink_set))

                if not M[xi] == M[xi] | set(xj_list):
                    M[xi] = M[xi] | set(xj_list)
                    changed = True

            hu_history[tuple(U)] = H_U

        if changed:
            l = 1
        elif l < max_explanatory_num:
            l += 1
        else:
            break

    return M


def get_common_ancestors(M, U):
    """Get the set of common ancestors of U"""
    Mj_list = [M[xj] for xj in U]
    return set.intersection(*Mj_list)


def get_residual_matrix(X, U, H_U):
    if len(H_U) == 0:
        return X

    Y = np.zeros_like(X)
    for xj in U:
        Y[:, xj], _ = get_resid_and_coef(X, xj, list(H_U))
    return Y


def is_non_gaussianity(Y, U, alpha):
    """Test whether a variable is generated from a non-Gaussian process using the Shapiro-Wilk test"""
    for xj in U:
        if shapiro(Y[:, xj])[1] > alpha:
            return False
    return True


def is_correlated(a, b, alpha):
    """Estimate that the two variables are linearly correlated using the Pearson's correlation"""
    return pearsonr(a, b)[1] < alpha


def exists_ancestor_in_U(M, U, xi, xj_list):
    # Check xi is not in Mj, the ancestor of xj.
    for xj in xj_list:
        if xi in M[xj]:
            return True

    # Check if xj_list is a subset of Mi, the ancestor of xi.
    if set(xj_list) == set(xj_list) & M[xi]:
        return True
    return False


def is_independent_of_resid(Y, xi, xj_list, alpha, bw_method, MLHSICR):
    """Check whether the residuals obtained from multiple regressions are independent"""
    n_samples = Y.shape[0]

    # Multiple Regression with OLS.
    is_all_independent = True
    resid, _ = get_resid_and_coef(Y, xi, xj_list)
    for xj in xj_list:
        if not is_independent(
            np.reshape(resid, [n_samples, 1]),
            np.reshape(Y[:, xj], [n_samples, 1]),
            bw_method,
            alpha,
        ):
            is_all_independent = False
            break

    if is_all_independent:
        return True
    elif len(xj_list) == 1 or MLHSICR is False:
        return False

    # Multiple Regression with MLHSICR.
    resid, _ = get_resid_and_coef_by_MLHSICR(Y, xi, xj_list)
    for xj in xj_list:
        if not is_independent(
            np.reshape(resid, [n_samples, 1]),
            np.reshape(Y[:, xj], [n_samples, 1]),
            bw_method,
            alpha,
        ):
            return False
    return True


def get_resid_and_coef(X, endog_idx, exog_idcs):
    """Get the residuals and coefficients of the ordinary least squares method"""
    lr = LinearRegression()
    lr.fit(X[:, exog_idcs], X[:, endog_idx])
    resid = X[:, endog_idx] - lr.predict(X[:, exog_idcs])
    return resid, lr.coef_


def get_resid_and_coef_by_MLHSICR(Y, xi, xj_list):
    """Get the residuals and coefficients by minimizing the sum of HSICs using the L-BFGS method."""
    n_samples = Y.shape[0]
    width_list = []
    Lc_list = []
    for xj in xj_list:
        yj = np.reshape(Y[:, xj], [n_samples, 1])
        width_xj = get_kernel_width(yj)
        _, Lc = get_gram_matrix(yj, width_xj)

        width_list.append(width_xj)
        Lc_list.append(Lc)

    _, initial_coef = get_resid_and_coef(Y, xi, xj_list)
    width_xi = get_kernel_width(np.reshape(Y[:, xi], [n_samples, 1]))

    # Calculate the sum of the Hilbert-Schmidt independence criterion
    def sum_empirical_hsic(coef):
        resid = Y[:, xi]
        width = width_xi
        for j, xj in enumerate(xj_list):
            resid = resid - coef[j] * Y[:, xj]
            width = width - coef[j] * width_list[j]
        _, Kc = get_gram_matrix(np.reshape(resid, [n_samples, 1]), width)

        objective = 0.0
        for j, xj in enumerate(xj_list):
            objective += hsic_teststat(Kc, Lc_list[j], n_samples)
        return objective

    # Estimate coefficients by minimizing the sum of HSICs using the L-BFGS method.
    coefs, _, _ = fmin_l_bfgs_b(
        func=sum_empirical_hsic, x0=initial_coef, approx_grad=True
    )

    resid = Y[:, xi]
    for j, xj in enumerate(xj_list):
        resid = resid - coefs[j] * Y[:, xj]
    return resid, coefs


def is_independent(X, Y, bw_method, alpha):
    _, p = hsic_test_gamma(X, Y, bw_method=bw_method)
    return p > alpha



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""


import copy
import itertools

import numpy as np
from pygam import LinearGAM
from sklearn.utils import check_array

from .hsic import hsic_test_gamma


class CAMUV:
    """Implementation of CAM-UV Algorithm [1]_

    References
    ----------
    .. [1] T.N.Maeda and S.Shimizu. Causal additive models with unobserved variables.
       In Proc. Thirty-Seventh Conference on Uncertainty in Artificial Intelligence (UAI). PMLR  161:97-106, 2021.
    """

    def __init__(self, alpha=0.01, num_explanatory_vals=2):
        """Construct a CAM-UV model.

        Parameters
        ----------
         alpha : float, optional (default=0.01)
             Alpha level.
         num_explanatory_vals : int, optional (default=2)
             Maximum number of explanatory variables.
        """

        # Check parameters
        if num_explanatory_vals <= 0:
            raise ValueError("num_explanatory_vals must be > 0.")

        if alpha < 0:
            raise ValueError("alpha must be >= 0.")

        self._num_explanatory_vals = num_explanatory_vals
        self._alpha = alpha

    def fit(self, X):
        X = check_array(X)

        n = X.shape[0]
        d = X.shape[1]
        N = self._get_neighborhoods(X, self._alpha)
        P = self._find_parents(X, self._alpha, self._num_explanatory_vals, N)

        U = []

        for i in range(d):
            for j in range(d)[i + 1 :]:
                if (i in P[j]) or (j in P[i]):
                    continue
                if (i not in N[j]) or (j not in N[i]):
                    continue
                i_residual = self._get_residual(X, i, P[i])
                j_residual = self._get_residual(X, j, P[j])
                in_X = np.reshape(i_residual, [n, 1])
                in_Y = np.reshape(j_residual, [n, 1])
                independence = hsic_test_gamma(X=in_X, Y=in_Y)[1]
                if independence < self._alpha:
                    if not set([i, j]) in U:
                        U.append(set([i, j]))

        return self._estimate_adjacency_matrix(X, P, U)

    def _get_residual(self, X, explained_i, explanatory_ids):
        explanatory_ids = list(explanatory_ids)

        if len(explanatory_ids) == 0:
            residual = X[:, explained_i]
        else:
            gam = LinearGAM().fit(X[:, explanatory_ids], X[:, explained_i])
            residual = X[:, explained_i] - gam.predict(X[:, explanatory_ids])
        return residual

    def _get_neighborhoods(self, X, alpha):
        n = X.shape[0]
        d = X.shape[1]
        N = [set() for i in range(d)]
        for i in range(d):
            for j in range(d)[i + 1 :]:
                in_X = np.reshape(X[:, i], [n, 1])
                in_Y = np.reshape(X[:, j], [n, 1])
                independence = hsic_test_gamma(X=in_X, Y=in_Y)[1]
                if independence < self._alpha:
                    N[i].add(j)
                    N[j].add(i)
        return N

    def _find_parents(self, X, alpha, maxnum_vals, N):
        n = X.shape[0]
        d = X.shape[1]
        P = [set() for i in range(d)]  # Parents
        t = 2
        Y = copy.deepcopy(X)

        while True:
            changed = False
            variables_set_list = list(itertools.combinations(set(range(d)), t))
            for variables_set in variables_set_list:
                variables_set = set(variables_set)

                if not self._check_identified_causality(variables_set, P):
                    continue

                child, independence_with_K = self._get_child(
                    X, variables_set, P, N, Y, alpha
                )
                if not independence_with_K > alpha:
                    continue

                parents = variables_set - {child}
                if not self._check_independence_withou_K(
                    parents, child, P, N, Y, alpha
                ):
                    continue

                for parent in parents:
                    P[child].add(parent)
                    changed = True
                    Y = self._get_residuals_matrix(X, Y, P, child)

            if changed:
                t = 2
            else:
                t += 1
                if t > maxnum_vals:
                    break

        for i in range(d):
            non_parents = set()
            for j in P[i]:
                residual_i = self._get_residual(X, i, P[i] - {j})
                residual_j = self._get_residual(X, j, P[j])
                in_X = np.reshape(residual_i, [n, 1])
                in_Y = np.reshape(residual_j, [n, 1])
                independence = hsic_test_gamma(X=in_X, Y=in_Y)[1]
                if independence > alpha:
                    non_parents.add(j)
            P[i] = P[i] - non_parents

        return P

    def _get_residuals_matrix(self, X, Y_old, P, child):
        Y = copy.deepcopy(Y_old)
        Y[:, child] = self._get_residual(X, child, P[child])
        return Y

    def _get_child(self, X, variables_set, P, N, Y, alpha):
        n = X.shape[0]

        max_independence = 0.0
        max_independence_child = None

        for child in variables_set:
            parents = variables_set - {child}

            if not self._check_correlation(child, parents, N):
                continue

            residual = self._get_residual(X, child, parents | P[child])
            in_X = np.reshape(residual, [n, 1])
            in_Y = np.reshape(Y[:, list(parents)], [n, len(parents)])
            independence = hsic_test_gamma(X=in_X, Y=in_Y)[1]
            if max_independence < independence:
                max_independence = independence
                max_independence_child = child

        return max_independence_child, max_independence

    def _check_independence_withou_K(self, parents, child, P, N, Y, alpha):
        n = Y.shape[0]
        for parent in parents:
            in_X = np.reshape(Y[:, child], [n, 1])
            in_Y = np.reshape(Y[:, parent], [n, 1])
            independence = hsic_test_gamma(X=in_X, Y=in_Y)[1]
            if alpha < independence:
                return False
        return True

    def _check_identified_causality(self, variables_set, P):
        variables_list = list(variables_set)
        for i in variables_list:
            for j in variables_list[variables_list.index(i) + 1 :]:
                if (j in P[i]) or (i in P[j]):
                    return False
        return True

    def _check_correlation(self, child, parents, N):
        for parent in parents:
            if parent not in N[child]:
                return False
        return True

    def _estimate_adjacency_matrix(self, X, P, U):
        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i, parents in enumerate(P):
            for parent in parents:
                B[i, parent] = 1
        for confounded_pair in U:
            confounded_pair = list(confounded_pair)
            B[confounded_pair[0], confounded_pair[1]] = np.nan
            B[confounded_pair[1], confounded_pair[0]] = np.nan
        self._adjacency_matrix = B
        return self

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrix



import os

import numpy as np
import pandas as pd

from lingam.direct_lingam import DirectLiNGAM
from lingam.varma_lingam import VARMALiNGAM

def randnetbalanced(dims, samples, indegree, parminmax, errminmax):
    """
    この関数は以前頂いたmatlabのスクリプトを移植したものですのでご確認不要です。
    create a more balanced random network

    Parameter
    ---------
    dims : int
        number of variables
    samples : int
        number of samples
    indegree : int or float('inf')
        number of parents of each node (float('inf') = fully connected)
    parminmax : dictionary
        standard deviation owing to parents 
    errminmax : dictionary
        standard deviation owing to error variable

    Return
    ------
    B : array, shape (dims, dims)
        the strictly lower triangular network matrix
    errstd : array, shape (dims, 1)
        the vector of error (disturbance) standard deviations
    """

    # First, generate errstd
    errstd = np.random.uniform(low=errminmax['min'], high=errminmax['max'], size=(dims, 1))

    # Initializations
    X = np.empty(shape=[dims, samples])
    B = np.zeros([dims, dims])

    # Go trough each node in turn
    for i in range(dims):

        # If indegree is finite, randomly pick that many parents,
        # else, all previous variables are parents
        if indegree == float('inf'):
            if i <= indegree:
                par = np.arange(i)
            else:
                par = np.random.permutation(i)[:indegree]
        else:
            par = np.arange(i)

        if len(par) == 0:
            # if node has no parents
            # Increase errstd to get it to roughly same variance
            parent_std = np.random.uniform(low=parminmax['min'], high=parminmax['max'])
            errstd[i] = np.sqrt(errstd[i]**2 + parent_std**2)

            # Set data matrix to empty
            X[i] = np.zeros(samples)
        else:
            # If node has parents, do the following
            w = np.random.normal(size=[1, len(par)])

            # Randomly pick weights
            wfull = np.zeros([1, i])
            wfull[0, par] = w

            # Calculate contribution of parents
            X[i] = np.dot(wfull, X[:i, :])

            # Randomly select a 'parents std' 
            parstd = np.random.uniform(low=parminmax['min'], high=parminmax['max'])

            # Scale w so that the combination of parents has 'parstd' std
            scaling = parstd / np.sqrt(np.mean(X[i] ** 2))
            w = w * scaling

            # Recalculate contribution of parents
            wfull = np.zeros([1, i])
            wfull[0, par] = w
            X[i] = np.dot(wfull, X[:i, :])

            # Fill in B
            B[i, par] = w

        # Update data matrix
        X[i] = X[i] + np.random.normal(size=samples) * errstd[i]

    return B, errstd

def generate_data(n=5, T=800, initial_data=None):
    head = 100
    T = T + head
    
    if initial_data is None:
        # psi0
        indegree = float('inf')
        psi0, _ = randnetbalanced(n, n, indegree, {'min':0.05, 'max':0.5}, {'min':0.05, 'max':0.5})
        permutation = np.random.permutation(n)
        psi0 = psi0[permutation][:, permutation]

        # causal order
        causal_order = np.empty(len(permutation))
        causal_order[permutation] = np.arange(len(permutation))
        causal_order = causal_order.astype(int)

        # phi1
        value = np.random.uniform(low=0.01, high=0.5, size=(n, n))
        sign = np.random.choice([-1, 1], size=(n, n))
        phi1 = np.multiply(value, sign)

        # theta1
        value = np.random.uniform(low=0.01, high=0.5, size=(n, n))
        sign = np.random.choice([-1, 1], size=(n, n))
        theta1 = np.multiply(value, sign)
    else:
        psi0 = initial_data['psi0']
        phi1 = initial_data['phi1']
        theta1 = initial_data['theta1']
        causal_order = initial_data['causal_order']

    # psi1, omega1
    psi1 = np.dot(np.eye(n) - psi0, phi1)
    omega1 = np.dot(np.eye(n) - psi0, theta1, np.linalg.inv(np.eye(n) - psi0))
                    
    # external influence
    expon = 0.1
    ext = np.empty((n, T))
    for i in range(n):
        ext[i, :] = np.random.normal(size=(1, T))
        ext[i, :] = np.multiply(np.sign(ext[i, :]), abs(ext[i, :]) ** expon)
        ext[i, :] = ext[i, :] - np.mean(ext[i, :])
        ext[i, :] = ext[i, :] / np.std(ext[i, :])

    # observed signals y
    y = np.zeros((n, T))
    y[:, 0] = np.random.normal(loc=0.1, scale=1, size=(n, )) * np.random.choice([-1, 1], size=(n, ))
    for t in range(1, T):
        for i in causal_order:
            y[i, t] = np.dot(psi0[i, :], y[:, t]) + np.dot(psi1[i, :], y[:, t - 1]) + ext[i, t] + np.dot(omega1[i, :], ext[:, t - 1])


    return y[:, head:].T, psi0, psi1, omega1, causal_order

def test_fit_success():
    X, psi0, phi1, omega1, causal_order = generate_data(n=3, T=30)

    # default
    model = VARMALiNGAM()
    model.fit(X)
    co = model.causal_order_
    am = model.adjacency_matrices_
    resid = model.residuals_
    p_values = model.get_error_independence_p_values()

    print('_ma_coefs:\n', model._ma_coefs)

    # order=(2, 1)
    model = VARMALiNGAM(order=(2, 2))
    model.fit(X)

    # criterion='aic'
    model = VARMALiNGAM(criterion='aic')
    model.fit(X)

    # prune=True
    model = VARMALiNGAM(prune=True)
    model.fit(X)

    model = VARMALiNGAM(prune=True, criterion=None, order=(2, 2))
    model.fit(X)

    # ar_coefs
    model = VARMALiNGAM(ar_coefs=[[[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]]])
    model.fit(X)

    # ma_coefs
    model = VARMALiNGAM(ma_coefs=[[[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]]])
    model.fit(X)

    # lingam_model
    m = DirectLiNGAM()
    model = VARMALiNGAM(lingam_model=m)
    model.fit(X)

def test_fit_invalid():
    X, psi0, phi1, omega1, causal_order = generate_data(n=3, T=30)

    # invalid lingam_model
    try:
        model = VARMALiNGAM(lingam_model=1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    X, psi0, phi1, omega1, causal_order = generate_data(n=3, T=30)

    model = VARMALiNGAM()
    result = model.bootstrap(X, n_sampling=3)

    result.adjacency_matrices_
    result.total_effects_

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    result.get_paths(1, 0, 1, 0)

def test_bootstrap_invalid():
    X, psi0, phi1, omega1, causal_order = generate_data(n=3, T=100)

    model = VARMALiNGAM()
    result = model.bootstrap(X, n_sampling=3)

    result.adjacency_matrices_
    result.total_effects_

    # min_causal_effect
    try:
        result.get_paths(1, 0, 0, 1, -1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # to_lag > from_lag
    try:
        result.get_paths(1, 0, 0, 1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # same variable
    try:
        result.get_paths(0, 0, 1, 1)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_estimate_total_effect_invalid():
    X, psi0, phi1, omega1, causal_order = generate_data(n=3, T=30)

    model = VARMALiNGAM()
    model.fit(X)
    model._causal_order = [0, 1, 2]

    # warning
    model.estimate_total_effect(X, model.residuals_, 2, 1)



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    README = fh.read()

import lingam

VERSION = lingam.__version__

setuptools.setup(
    name='lingam',
    version=VERSION,
    author='T.Ikeuchi, G.Haraoka, M.Ide, Y.Zeng, T.N.Maeda, W.Kurebayashi, S.Shimizu', 
    description='LiNGAM Python Package',
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'graphviz',
        'statsmodels',
        'networkx',
        'pandas',
        'pygam',
        'matplotlib',
        'psy',
    ],
    url='https://github.com/cdt15/lingam',
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)



import os

import numpy as np
import pandas as pd

from lingam.direct_lingam import DirectLiNGAM
from lingam.var_lingam import VARLiNGAM


def generate_data(n=5, T=1000, random_state=None, initial_data=None):
    """
    Parameter
    ---------
    n : int
        number of variables
    T : int
        number of samples
    random_state : int
        seed for np.random.seed
    initial_data : list of np.ndarray
        dictionary of initial datas
    """

    T_spurious = 20
    expon = 1.5
    
    if initial_data is None:
        permutation = np.random.permutation(n)
        
        value = np.random.uniform(low=0.05, high=0.5, size=(n, n))
        sign = np.random.choice([-1, 1], size=(n, n))
        B0 = np.multiply(value, sign)
        
        B0 = np.multiply(B0, np.random.binomial(1, 0.4, size=(n, n)))
        B0 = np.tril(B0, k=-1)
        B0 = B0[permutation][:, permutation]

        value = np.random.uniform(low=0.05, high=0.5, size=(n, n))
        sign = np.random.choice([-1, 1], size=(n, n))
        B1 = np.multiply(value, sign)
        B1 = np.multiply(B1, np.random.binomial(1, 0.4, size=(n, n)))
        
        causal_order = np.empty(len(permutation))
        causal_order[permutation] = np.arange(len(permutation))
        causal_order = causal_order.astype(int)
    else:
        B0 = initial_data['B0']
        B1 = initial_data['B1']
        causal_order =initial_data['causal_order'] 
        
    M1 = np.dot(np.linalg.inv(np.eye(n) - B0), B1);

    ee = np.empty((n, T + T_spurious))
    for i in range(n):
        ee[i, :] = np.random.normal(size=(1, T + T_spurious));
        ee[i, :] = np.multiply(np.sign(ee[i, :]), abs(ee[i, :]) ** expon);
        ee[i, :] = ee[i, :] - np.mean(ee[i, :]);
        ee[i, :] = ee[i, :] / np.std(ee[i, :]);

    std_e = np.random.uniform(size=(n,)) + 0.5
    nn = np.dot(np.dot(np.linalg.inv(np.eye(n) - B0), np.diag(std_e)), ee);

    xx = np.zeros((n, T + T_spurious))
    xx[:, 0] = np.random.normal(size=(n, ));

    for t in range(1, T + T_spurious):
        xx[:, t] = np.dot(M1, xx[:, t - 1]) + nn[:, t];

    data = xx[:, T_spurious + 1 : T_spurious + T];
    
    return data.T, B0, B1, causal_order

def test_fit_success():
    X, B0, B1, causal_order = generate_data(n=3, T=100)

    # default
    model = VARLiNGAM()
    model.fit(X)
    co = model.causal_order_
    am = model.adjacency_matrices_
    resid = model.residuals_
    p_values = model.get_error_independence_p_values()

    # lags=2
    model = VARLiNGAM(lags=2)
    model.fit(X)

    # criterion='aic'
    model = VARLiNGAM(criterion='aic')
    model.fit(X)

    # prune=True
    model = VARLiNGAM(prune=True)
    model.fit(X)

    # ar_coefs
    model = VARLiNGAM(ar_coefs=[[[0.1, 0.1, 0.1], [0.1, 0.1, 0.1], [0.1, 0.1, 0.1]]])
    model.fit(X)

    # lingam_model
    m = DirectLiNGAM()
    model = VARLiNGAM(lingam_model=m)
    model.fit(X)

def test_fit_invalid():
    X, B0, B1, causal_order = generate_data(n=3, T=100)

    # invalid lingam_model
    try:
        model = VARLiNGAM(lingam_model=1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    X, B0, B1, causal_order = generate_data(n=3, T=100)

    model = VARLiNGAM()
    result = model.bootstrap(X, n_sampling=3)

    result.adjacency_matrices_
    result.total_effects_

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    result.get_paths(1, 0, 1, 0)

def test_bootstrap_invalid():
    X, B0, B1, causal_order = generate_data(n=3, T=100)

    model = VARLiNGAM()
    result = model.bootstrap(X, n_sampling=3)

    result.adjacency_matrices_
    result.total_effects_

    # min_causal_effect
    try:
        result.get_paths(1, 0, 0, 1, -1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # to_lag > from_lag
    try:
        result.get_paths(1, 0, 0, 1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # same variable
    try:
        result.get_paths(0, 0, 1, 1)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_estimate_total_effect_invalid():
    X, B0, B1, causal_order = generate_data(n=3, T=100)

    model = VARLiNGAM()
    model.fit(X)
    model._causal_order = [0, 1, 2]

    # warning
    model.estimate_total_effect(X, 2, 1)
